// OS X port (C) 2003 Eli Curtz <eli@nuprometheus.com>
// Original Linux version (C) 2003 Ulrich Hecht <uli@emulinks.de>
// Licensed under the terms of the GNU Public License version 2
// Portions of this file were based on the Apple USB EZLoader sample code

#include <unistd.h>
#include <mach/mach.h>

#include <CoreFoundation/CoreFoundation.h>

#include <IOKit/IOKitLib.h>
#include <IOKit/IOCFPlugIn.h>
#include <IOKit/usb/IOUSBLib.h>

#include "comm2.h"

// constants
#define	k8051_USBCS		0x7f92

#define kF2AOrigVendorID	0x0547
#define kF2AOrigProductID	0x2131
#define kF2AConfigVendorID	0x0547
#define kF2AConfigProductID	0x1002

#define kUSBConnectionTimeout	5.0
#define kUSBSendTimeout		500
#define kUSBReceiveTimeout	1000

// globals
IONotificationPortRef	gNotifyPort;
IOUSBDeviceInterface**	gNewDevice = NULL;

io_iterator_t		gRawAddedIter;
io_iterator_t		gRawRemovedIter;
io_iterator_t		gNewDeviceAddedIter;
io_iterator_t		gNewDeviceRemovedIter;

extern UInt8 gVerbosity;


void SignalHandler(int sigraised)
{
    printf("\nInterrupted\n");
   
    // Clean up here
    CleanupCommunication();

    // exit(0) should not be called from a signal handler.  Use _exit(0) instead
    _exit(0);
}

IOReturn EstablishCommunication()
{
    mach_port_t 		masterPort;
    kern_return_t		kr;
    CFMutableDictionaryRef 	matchingDict;
    CFRunLoopRef		runLoop;
    CFRunLoopSourceRef		runLoopSource;
    CFStringRef 		runLoopMode;
    SInt32			runLoopResult;
    SInt32			usbVendor = kF2AOrigVendorID;
    SInt32			usbProduct = kF2AOrigProductID;
    SInt32			usbVendor2 = kF2AConfigVendorID;
    SInt32			usbProduct2 = kF2AConfigProductID;
    sig_t			oldHandler;

    // Set up a signal handler so we can clean up when we're interrupted from the command line
    // Otherwise we stay in our run loop forever.
    oldHandler = signal(SIGINT, SignalHandler);
    if (oldHandler == SIG_ERR)
        printf("WARNING: Could not establish new signal handler");
        
    // First create a master_port for the task
    kr = IOMasterPort(MACH_PORT_NULL, &masterPort);
    if ((kr != kIOReturnSuccess) || (masterPort == NULL))
    {
        printf("ERR: Couldn't create a master IOKit Port(%08x)\n", kr);
        return -1;
    }

    if (gVerbosity > 1)
    {
        printf("\nLooking for devices matching vendor ID=%ld and product ID=%ld\n", usbVendor, usbProduct);
    }

    // Set up the matching criteria for instances of class IOUSBDevice and its subclasses
    matchingDict = IOServiceMatching(kIOUSBDeviceClassName);
    if (matchingDict == NULL)
    {
        printf("ERR: Couldn't create a USB matching dictionary\n");
        mach_port_deallocate(mach_task_self(), masterPort);
        return -1;
    }
    
    // Add our vendor and product IDs to the matching criteria
    CFDictionarySetValue( 
            matchingDict, 
            CFSTR(kUSBVendorID), 
            CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &usbVendor)); 
    CFDictionarySetValue( 
            matchingDict, 
            CFSTR(kUSBProductID), 
            CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &usbProduct)); 

    // Create a notification port and add its run loop event source to our run loop
    // This is how async notifications get set up.
    gNotifyPort = IONotificationPortCreate(masterPort);
    runLoop = CFRunLoopGetCurrent();
    runLoopSource = IONotificationPortGetRunLoopSource(gNotifyPort);
    runLoopMode = CFStringCreateWithCString(kCFAllocatorDefault, "F2Ainit", kCFStringEncodingASCII);
    
    CFRunLoopAddSource(runLoop, runLoopSource, runLoopMode);
    
    // Retain additional references because we use this same dictionary with four calls to 
    // IOServiceAddMatchingNotification, each of which consumes one reference.
    matchingDict = (CFMutableDictionaryRef) CFRetain( matchingDict ); 
    matchingDict = (CFMutableDictionaryRef) CFRetain( matchingDict ); 
    matchingDict = (CFMutableDictionaryRef) CFRetain( matchingDict ); 
    
    // Now set up two notifications, one to be called when a raw device is first matched by I/O Kit,
    // and the other to be called when the device is terminated.
    kr = IOServiceAddMatchingNotification(  gNotifyPort,
                                            kIOFirstMatchNotification,
                                            matchingDict,
                                            RawDeviceAdded,
                                            NULL,
                                            &gRawAddedIter );
                                            
    // Iterate once to get already-present devices and arm the notification
    RawDeviceAdded(NULL, gRawAddedIter);
    
    kr = IOServiceAddMatchingNotification(  gNotifyPort,
                                            kIOTerminatedNotification,
                                            matchingDict,
                                            RawDeviceRemoved,
                                            NULL,
                                            &gRawRemovedIter );
                                            
    // Iterate once to arm the notification
    RawDeviceRemoved(NULL, gRawRemovedIter);	
    
    // Change the USB product ID in our matching dictionary to the one the 
    // device will have once the firmware has been downloaded.
    if (gVerbosity > 1)
    {
        printf("Configured devices will have vendor ID=%ld and product ID=%ld\n", usbVendor2, usbProduct2);
    }
    
    CFDictionarySetValue( 
            matchingDict, 
            CFSTR(kUSBVendorID), 
            CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &usbVendor2)); 
    CFDictionarySetValue( 
            matchingDict, 
            CFSTR(kUSBProductID), 
            CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &usbProduct2)); 

    // Now set up two more notifications, one to be called when a new test device is first matched
    // by I/O Kit, and the other to be called when the device is terminated.
    kr = IOServiceAddMatchingNotification(  gNotifyPort,
                                            kIOFirstMatchNotification,
                                            matchingDict,
                                            NewDeviceAdded,
                                            NULL,
                                            &gNewDeviceAddedIter );
                                            
    // Iterate once to get already-present devices and arm the notification
    NewDeviceAdded(NULL, gNewDeviceAddedIter);

    kr = IOServiceAddMatchingNotification(  gNotifyPort,
                                            kIOTerminatedNotification,
                                            matchingDict,
                                            NewDeviceRemoved,
                                            NULL,
                                            &gNewDeviceRemovedIter );
                                            
    // Iterate once to arm the notification
    NewDeviceRemoved(NULL, gNewDeviceRemovedIter); 	

    // Done with the master_port
    mach_port_deallocate(mach_task_self(), masterPort);
    masterPort = 0;

    // If we already found the device, no need for notification loop
    if (gNewDevice != NULL)
        return kIOReturnSuccess;

    // Start the run loop. Now we'll receive notifications.
    runLoopResult = CFRunLoopRunInMode(runLoopMode, kUSBConnectionTimeout, false);
    
    // If we find a configured service it will stop the runLoop
    if (runLoopResult == kCFRunLoopRunStopped)
        return kIOReturnSuccess;
    // Timed out, probably because there wasn't a device attached
    if (runLoopResult == kCFRunLoopRunTimedOut)
        return kIOReturnNoDevice;
    // Shouldn't get these
    if (runLoopResult == kCFRunLoopRunFinished)
        return kIOReturnError;
    if (runLoopResult == kCFRunLoopRunHandledSource)
        return kIOReturnError;
    
    // Won't get here unless CFRunLoopRunInMode() returns an unexpected result    
    return kIOReturnError;
}

void CleanupCommunication()
{
    IONotificationPortDestroy(gNotifyPort);
    
    if (gNewDevice)
    {
        (*gNewDevice)->USBDeviceClose(gNewDevice);
        (*gNewDevice)->Release(gNewDevice);
    }

    if (gRawAddedIter) 
    {
        IOObjectRelease(gRawAddedIter);
        gRawAddedIter = 0;
    }

    if (gRawRemovedIter) 
    {
        IOObjectRelease(gRawRemovedIter);
        gRawRemovedIter = 0;
    }
    
    if (gNewDeviceAddedIter) 
    {
        IOObjectRelease(gNewDeviceAddedIter);
        gNewDeviceAddedIter = 0;
    }

    if (gNewDeviceRemovedIter) 
    {
        IOObjectRelease(gNewDeviceRemovedIter);
        gNewDeviceRemovedIter = 0;
    }
}

IOReturn ConfigureAnchorDevice(IOUSBDeviceInterface **dev)
{
    UInt8				numConf;
    IOReturn				kernResult;
    IOUSBConfigurationDescriptorPtr	confDesc;
    
    kernResult = (*dev)->GetNumberOfConfigurations(dev, &numConf);
    if (!numConf)
        return kIOReturnError;
    
    // Get the configuration descriptor for index 0
    kernResult = (*dev)->GetConfigurationDescriptorPtr(dev, 0, &confDesc);
    if (kernResult != kIOReturnSuccess)
    {
        printf("USB Error: Unable to get config descriptor for index 0 (%08x)\n", kernResult);
        return kernResult;
    }
    kernResult = (*dev)->SetConfiguration(dev, confDesc->bConfigurationValue);
    if (kernResult != kIOReturnSuccess)
    {
        printf("USB Error: Unable to set configuration (%08x)\n", kernResult);
        return kernResult;
    }
    
    return kIOReturnSuccess;
}

IOReturn AnchorWrite(IOUSBDeviceInterface **dev, UInt16 anchorAddress, UInt16 count, UInt8 writeBuffer[])
{
    IOUSBDevRequest 		request;
    
    request.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBVendor, kUSBDevice);
    request.bRequest = 0xa0;
    request.wValue = anchorAddress;
    request.wIndex = 0;
    request.wLength = count;
    request.pData = writeBuffer;

    return (*dev)->DeviceRequest(dev, &request);
}


FILE *openFile(UInt16 vendor, UInt16 product)
{
char name[]="f2afirm.hex";
char path[256];
char *pathp = path;
FILE *hexFile;

    pathp = getenv("HEXPATH");
    if(pathp != nil)
    {
        strncpy(path, pathp, 256);
    }
    else
    {
        pathp = getenv("PATH");
        strncpy(path, pathp, 256);
    }
//    printf("Found path %s\n", path);
    
//    sprintf(name, "%04X%04X.HEX", vendor, product);
    printf("File to open is \"%s\"\n", name);
    do{
        hexFile = fopen(name, "r");
        if(hexFile != nil)
        {
            getcwd(path, 255);
            printf("From directory %s\n", path);
            break;
        }
        
        if(pathp == nil)
        {
            break;
        }
        if( chdir(strsep(&pathp, ":")) != 0)
        {
            printf("chdir failed with errno %d\n", errno);
            perror(nil);
            break;
        }

    }while(1);
    
    return(hexFile);
}

/*
    Description of Intel hex records.

Position	Description
1	Record Marker: The first character of the line is always a colon (ASCII 0x3A) to identify 
                        the line as an Intel HEX file
2 - 3	Record Length: This field contains the number of data bytes in the register represented 
                        as a 2-digit hexidecimal number. This is the total number of data bytes, 
                        not including the checksum byte nor the first 9 characters of the line.
4 - 7	Address: This field contains the address where the data should be loaded into the chip. 
                        This is a value from 0 to 65,535 represented as a 4-digit hexidecimal value.
8 - 9	Record Type: This field indicates the type of record for this line. The possible values 
                        are: 00=Register contains normal data. 01=End of File. 02=Extended address.
10 - ?	Data Bytes: The following bytes are the actual data that will be burned into the EPROM. The 
                        data is represented as 2-digit hexidecimal values.
Last 2 characters	Checksum: The last two characters of the line are a checksum for the line. The 
                        checksum value is calculated by taking the twos complement of the sum of all 
                        the preceeding data bytes, excluding the checksum byte itself and the colon 
                        at the beginning of the line.
*/

IOReturn hexRead(INTEL_HEX_RECORD *record, FILE *hexFile)
{	// Read the next hex record from the file into the structure

    // **** Need to impliment checksum checking ****
    
    
char c;
UInt16 i;
int n, c1, check, len;
    c = getc(hexFile);
    
    if(c != ':')
    {
        printf("Line does not start with colon (%d)\n", c);
        return(kIOReturnNotAligned);
    }
    n = fscanf(hexFile, "%2lX%4lX%2lX", &record->Length, &record->Address, &record->Type);
    if(n != 3)
    {
        printf("Could not read line preamble %d\n", c);
        return(kIOReturnNotAligned);
    }
    
    len = record->Length;
    if(len > MAX_INTEL_HEX_RECORD_LENGTH)
    {
        printf("length is more than can fit %d, %d\n", len, MAX_INTEL_HEX_RECORD_LENGTH);
        return(kIOReturnNotAligned);
   }
    for(i = 0; i<len; i++)
    {
        n = fscanf(hexFile, "%2X", &c1);
        if(n != 1)
        {
            if(i != record->Length)
            {
                printf("Line finished at wrong time %d, %ld\n", i, record->Length);
                return(kIOReturnNotAligned);
            }
        }
        record->Data[i] = c1;
    
    }
    n = fscanf(hexFile, "%2X\n", &check);
    if(n != 1)
    {
        printf("Check not found\n");
        return(kIOReturnNotAligned);
    }
    
    return kIOReturnSuccess;
}

IOReturn DownloadToAnchorDevice(IOUSBDeviceInterface **dev, UInt16 vendor, UInt16 product)
{
    UInt8 	writeVal;
    IOReturn	kr;
FILE *hexFile;
INTEL_HEX_RECORD anchorCode;
    
    hexFile = openFile(vendor, product);
    if(hexFile == nil)
    {
        printf("File open failed\n");
        return(kIOReturnNotOpen);
    }

    // Assert reset
    writeVal = 1;
    kr = AnchorWrite(dev, k8051_USBCS, 1, &writeVal);
    if (kIOReturnSuccess != kr) 
    {
        printf("AnchorWrite reset returned err 0x%x!\n", kr);
        (*dev)->USBDeviceClose(dev);
        (*dev)->Release(dev);
        return kr;
    }
    
    // Download code
    while (1) 
    {
        kr = hexRead(&anchorCode, hexFile);
        if(anchorCode.Type != 0)
        {
            break;
        }
        if(kr == kIOReturnSuccess)
        {
            kr = AnchorWrite(dev, anchorCode.Address, anchorCode.Length, anchorCode.Data);
        }
        if (kIOReturnSuccess != kr) 
        {
            printf("AnchorWrite download %lx returned err 0x%x!\n", anchorCode.Address, kr);
            (*dev)->USBDeviceClose(dev);
            (*dev)->Release(dev);
            return kr;
        }
    }
    
    // De-assert reset
    writeVal = 0;
    kr = AnchorWrite(dev, k8051_USBCS, 1, &writeVal);
    if (kIOReturnSuccess != kr) 
    {
        printf("AnchorWrite run returned err 0x%x!\n", kr);
    }
    
    return kr;
}
      
void RawDeviceAdded(void *refCon, io_iterator_t iterator)
{
    kern_return_t		kr;
    io_service_t		usbDevice;
    IOCFPlugInInterface 	**plugInInterface=NULL;
    IOUSBDeviceInterface 	**dev=NULL;
    HRESULT 			res;
    SInt32 			score;
    UInt16			vendor;
    UInt16			product;
    UInt16			release;
    int exclusiveErr = 0;
    
    while ( (usbDevice = IOIteratorNext(iterator)) )
    {
        printf("Raw device added\n");
       
        kr = IOCreatePlugInInterfaceForService(usbDevice, kIOUSBDeviceUserClientTypeID, kIOCFPlugInInterfaceID, &plugInInterface, &score);
        kr = IOObjectRelease(usbDevice);				// done with the device object now that I have the plugin
        if ((kIOReturnSuccess != kr) || !plugInInterface)
        {
            printf("unable to create a plugin (%08x)\n", kr);
            continue;
        }
            
        // I have the device plugin, I need the device interface
        res = (*plugInInterface)->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes(kIOUSBDeviceInterfaceID), (LPVOID)&dev);
        (*plugInInterface)->Release(plugInInterface);			// done with this
        if (res || !dev)
        {
            printf("couldn't create a device interface (%08x)\n", (int) res);
            continue;
        }
        // technically should check these kr values
        kr = (*dev)->GetDeviceVendor(dev, &vendor);
        kr = (*dev)->GetDeviceProduct(dev, &product);
        kr = (*dev)->GetDeviceReleaseNumber(dev, &release);

        // need to open the device in order to change its state
        do{
            kr = (*dev)->USBDeviceOpen(dev);
            if(kIOReturnExclusiveAccess == kr)
            {
                exclusiveErr++;
                printf("Exclusive access err, sleeping on it %d\n", exclusiveErr);
                sleep(1);
            }
        }while( (kIOReturnExclusiveAccess == kr) && (exclusiveErr < 5) );
        
        if (kIOReturnSuccess != kr)
        {
            printf("unable to open device: %08x\n", kr);
            (void) (*dev)->Release(dev);
            continue;
        }
        kr = ConfigureAnchorDevice(dev);
        if (kIOReturnSuccess != kr)
        {
            printf("unable to configure device: %08x\n", kr);
            (void) (*dev)->USBDeviceClose(dev);
            (void) (*dev)->Release(dev);
            continue;
        }

        kr = DownloadToAnchorDevice(dev, vendor, product);
        if (kIOReturnSuccess != kr)
        {
            printf("unable to download to device: %08x\n", kr);
            (void) (*dev)->USBDeviceClose(dev);
            (void) (*dev)->Release(dev);
            continue;
        }

        kr = (*dev)->USBDeviceClose(dev);
        kr = (*dev)->Release(dev);
    }
}

void RawDeviceRemoved(void *refCon, io_iterator_t iterator)
{
    kern_return_t	kr;
    io_service_t	obj;
    
    while ( (obj = IOIteratorNext(iterator)) )
    {
        printf("Raw device removed\n");
        kr = IOObjectRelease(obj);
    }
}

void NewDeviceAdded(void *refCon, io_iterator_t iterator)
{
    kern_return_t		kr;
    io_service_t		usbDevice;
    CFRunLoopRef		runLoop;
    IOCFPlugInInterface 	**plugInInterface = NULL;
    HRESULT 			res;
    SInt32 			score;
    UInt16			vendor;
    UInt16			product;

    usbDevice = IOIteratorNext(iterator);
    if (usbDevice != NULL)
    {
        printf("New device added.\n");
        kr = IOCreatePlugInInterfaceForService(usbDevice, kIOUSBDeviceUserClientTypeID, kIOCFPlugInInterfaceID, &plugInInterface, &score);
        kr = IOObjectRelease(usbDevice);				// done with the device object now that I have the plugin
        if ((kIOReturnSuccess != kr) || !plugInInterface)
        {
            printf("unable to create a plugin (%08x)\n", kr);
            return;
        }
            
        // I have the device plugin, I need the device interface
        res = (*plugInInterface)->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes(kIOUSBDeviceInterfaceID), (LPVOID)&gNewDevice);
        (*plugInInterface)->Release(plugInInterface);			// done with this
        if (res || !gNewDevice)
        {
            printf("couldn't create a device interface (%08x)\n", (int) res);
            return;
        }
        // technically should check these kr values
        kr = (*gNewDevice)->GetDeviceVendor(gNewDevice, &vendor);
        kr = (*gNewDevice)->GetDeviceProduct(gNewDevice, &product);

        CFRunLoopStop(CFRunLoopGetCurrent());
    }
}

void NewDeviceRemoved(void *refCon, io_iterator_t iterator)
{
    kern_return_t	kr;
    io_service_t	obj;
    
    while ( (obj = IOIteratorNext(iterator)) )
    {
        printf("New device removed.\n");
        kr = IOObjectRelease(obj);
    }
}
