// OS X port (C) 2003 Eli Curtz <eli@nuprometheus.com>
// Original Linux version (C) 2003 Ulrich Hecht <uli@emulinks.de>
// Licensed under the terms of the GNU Public License version 2
// Portions of this file were based on the Apple USB EZLoader sample code

#include <unistd.h>

#include <CoreFoundation/CoreFoundation.h>

#include <IOKit/IOKitLib.h>
#include <IOKit/IOCFPlugIn.h>
#include <IOKit/usb/IOUSBLib.h>


// constants
#define kUSBSendTimeout		500
#define kUSBReceiveTimeout	1000

// globals
IOUSBInterfaceInterface182** gInterface;
extern IOUSBDeviceInterface** gNewDevice;

extern UInt8 gVerbosity;

IOReturn usb_connect()
{
    IOReturn			kernResult;
    IOUSBFindInterfaceRequest   request;
    io_iterator_t               iterator;
    io_service_t                usbInterface;
    IOCFPlugInInterface         **plugInInterface = NULL;
    HRESULT                     result;
    SInt32                      score;
    UInt8                       interfaceClass;
    UInt8                       interfaceSubClass;
        
    // Placing the constant kIOUSBFindInterfaceDontCare into the following
    // fields of the IOUSBFindInterfaceRequest structure will allow you
    // to find all the interfaces
    request.bInterfaceClass = kIOUSBFindInterfaceDontCare;
    request.bInterfaceSubClass = kIOUSBFindInterfaceDontCare;
    request.bInterfaceProtocol = kIOUSBFindInterfaceDontCare;
    request.bAlternateSetting = kIOUSBFindInterfaceDontCare;

    // Get an iterator for the interfaces on the device
    kernResult = (*gNewDevice)->CreateInterfaceIterator(gNewDevice, &request, &iterator);
    usbInterface = IOIteratorNext(iterator);
    if (usbInterface)
    {
        // Create an intermediate plug-in
        kernResult = IOCreatePlugInInterfaceForService(usbInterface,
                            kIOUSBInterfaceUserClientTypeID,
                            kIOCFPlugInInterfaceID,
                            &plugInInterface, &score);
                            
        // No longer need the usbInterface object after getting the plug-in
        kernResult = IOObjectRelease(usbInterface);
        if ((kernResult != kIOReturnSuccess) || !plugInInterface)
        {
            printf("USB Error: Unable to create a plug-in (%08x)\n", kernResult);
            return kernResult;
        }
        
        // Now create the device interface for the interface
        result = (*plugInInterface)->QueryInterface(plugInInterface,
                    CFUUIDGetUUIDBytes(kIOUSBInterfaceInterfaceID182),
                    (LPVOID) &gInterface);
                    
        // No longer need the intermediate plug-in
        (*plugInInterface)->Release(plugInInterface);

        if ((result != 0) || !gInterface)
        {
            printf("USB Error: Unable to create device interface (%08x)\n", (int) result);
            return kernResult;
        }
             
        // Show interface class and subclass
        if (gVerbosity > 0)
        {
            kernResult = (*gInterface)->GetInterfaceClass(gInterface, &interfaceClass);
            kernResult = (*gInterface)->GetInterfaceSubClass(gInterface, &interfaceSubClass);
    
            printf("Opened USB Interface to class %d, subclass %d\n", interfaceClass, interfaceSubClass);
        }

        // Now open the interface. This will cause the pipes associated with
        // the endpoints in the interface descriptor to be instantiated
        kernResult = (*gInterface)->USBInterfaceOpen(gInterface);
        if (kernResult != kIOReturnSuccess)
        {
            printf("USB Error: Unable to open interface (%08x)\n", kernResult);
            (void) (*gInterface)->Release(gInterface);
            gInterface = NULL;
            return kernResult;
        }
    }
            
    return kernResult;
}

void usb_disconnect()
{
    if (gVerbosity > 0)
    {
        printf("Closing USB Interface\n");
    }
    
    // Close up the interface
    if (gInterface)
    {
        (void) (*gInterface)->USBInterfaceClose(gInterface);
        (void) (*gInterface)->Release(gInterface);
        gInterface = NULL;
    }
}

IOReturn usb_read(char *buffer, UInt32 buffer_size)
{
    IOReturn kernResult;
    
    
    // Check for the interface
    if (!gInterface)
    {
        printf("USB Error: Attempted to read from missing interface\n");
    	return kIOReturnError;
    }
    
    // Read from the bulk input pipe
    kernResult = (*gInterface)->ReadPipeTO(gInterface, 3, buffer, &buffer_size, kUSBReceiveTimeout, kUSBReceiveTimeout);
    if (kernResult != kIOReturnSuccess)
    {
        printf("USB Error: Unable to perform bulk read (%08x)\n", kernResult);
        usb_disconnect();
    }
    
    return kernResult;
}

IOReturn usb_write(char *buffer, UInt32 buffer_size)
{
    IOReturn kernResult;
    
    // Check for the interface
    if (!gInterface)
    {
        printf("USB Error: Attempted to write to missing interface\n");
    	return kIOReturnError;
    }
    
    // Write to the bulk output pipe
    kernResult = (*gInterface)->WritePipeTO(gInterface, 4, buffer, buffer_size, kUSBSendTimeout, kUSBSendTimeout);
    if (kernResult != kIOReturnSuccess)
    {
        printf("USB Error: Unable to perform bulk write (%08x)\n", kernResult);
        usb_disconnect();
    }
    
    return kernResult;
}

