// OS X port (C) 2003 Eli Curtz <eli@nuprometheus.com>
// Original Linux version (C) 2003 Ulrich Hecht <uli@emulinks.de>
// Licensed under the terms of the GNU Public License version 2

#include <unistd.h>

#include <CoreFoundation/CoreFoundation.h>

#include <IOKit/IOKitLib.h>

#include "flash2a.h"


// constants
#define kMaxMultibootSize	256*1024

// globals
extern UInt8 gVerbosity;

IOReturn f2a_sendMessage(sendmsg* msg)
{
    // Convert the message parameters to little endian
    msg->command = CFSwapInt32HostToLittle(msg->command);
    msg->size = CFSwapInt32HostToLittle(msg->size);
    msg->_reserved1[0] = CFSwapInt32HostToLittle(msg->_reserved1[0]);
    msg->_reserved1[1] = CFSwapInt32HostToLittle(msg->_reserved1[1]);
    msg->magic = CFSwapInt32HostToLittle(msg->magic);
    msg->_reserved2[0] = CFSwapInt32HostToLittle(msg->_reserved2[0]);
    msg->_reserved2[1] = CFSwapInt32HostToLittle(msg->_reserved2[1]);
    msg->_reserved2[2] = CFSwapInt32HostToLittle(msg->_reserved2[2]);
    msg->unknown = CFSwapInt32HostToLittle(msg->unknown);
    msg->address = CFSwapInt32HostToLittle(msg->address);
    msg->sizekb = CFSwapInt32HostToLittle(msg->sizekb);
    // What about _reserved3? No idea what format it actually is...
    // unsigned char _reserved3[5*4-1]
    
    // Send the message
    return usb_write((char*) msg, sizeof(sendmsg));
}

IOReturn f2a_receiveMessage(recvmsg* msg)
{
    IOReturn kernResult;
    
    // Get the message
    kernResult = usb_read((char*) msg, sizeof(recvmsg));
    
    // If we knew the format we would want to convert to big endian
    
    return kernResult;
}

IOReturn f2a_boot(FILE* mbFile)
{
    IOReturn	kernResult;

    UInt32	fileSize;
    UInt32	fileKB;
    size_t	bytesRead;

    sendmsg sm;
    recvmsg rm;
    
    // Uli has the size set to 16k, which is smaller than
    // the multiboot file I have. Does the Windows version
    // not load the entire file?
    char fileBuffer[kMaxMultibootSize];
    memset(fileBuffer, 0, kMaxMultibootSize);

    fseek(mbFile, 0, SEEK_END);       
    fileSize = ftell(mbFile);
    fseek(mbFile, 0, SEEK_SET);       

    fileKB = fileSize / 1024;
    if (fileSize & (1024 - 1))
        fileKB++;
    
    // Read the multiboot data and close the file    
    bytesRead = fread(fileBuffer, 1, fileSize, mbFile);
    if (ferror(mbFile))
    {
        printf("File Error: Could not read boot file (%d)\n", ferror(mbFile));
    	return kIOReturnError;
    }

    // Wait for them to turn off the GBA. I'm not
    // sure why it has problems if the GBA is on during
    // firmware laoding, but something is not happy.
    printf("Please restart GBA in multiboot mode\n");
    do
    {
        sleep(1);
        kernResult = f2a_info(&rm);
    } while(rm.data[0] != 0);
    
    // Wait for a GBA in multiboot mode to be connected
    do
    {
        sleep(1);
        kernResult = f2a_info(&rm);
    } while(rm.data[0] == 0);

    // Boot the GBA
    memset(&sm, 0, sizeof(sendmsg));
    sm.command = CMD_MULTIBOOT1;

    kernResult = f2a_sendMessage(&sm);

    // Wait a few seconds (for the gba to start receiving?)
    sleep(5);

    memset(&sm, 0, sizeof(sendmsg));
    sm.command = CMD_MULTIBOOT2;
    sm.size = fileSize;

    kernResult = f2a_sendMessage(&sm);

    // Download the multiboot image
    sleep(3);

    UInt32 i;
    for(i = 0; i < fileKB; i++)
    {
        kernResult = usb_write(&fileBuffer[i*1024], 1024);
        sleep(0);
    }
    
    return kernResult;
}

IOReturn f2a_info(recvmsg* rm)
{
    IOReturn kernResult;
    sendmsg sm;
    
    memset(&sm, 0, sizeof(sendmsg));
    memset(rm, 0, sizeof(recvmsg));

    sm.command = CMD_GETINF;
        
    kernResult = f2a_sendMessage(&sm);
    kernResult = f2a_receiveMessage(rm);
    if (kernResult != kIOReturnSuccess)
    {
        printf("F2A Error: Couldn't read status info\n");
        return kernResult;
    }
    
    // Show the status of the F2A if verbosity is high
    if (gVerbosity > 1)
    {
        printf("info:");
        UInt32 i;
        for(i = 0; i < (sizeof(recvmsg)/sizeof(int)); i++)
                printf(" %08X",*(((unsigned int*)(rm))+i));
        puts("");
    }

    return kernResult;
}

IOReturn f2a_writeMultiboot(FILE* file)
{
    IOReturn 	kernResult;
    sendmsg 	sm;
    
    UInt32 	fileSize;
    char 	fileBuffer[1024];
    size_t 	bytesRead;
    
    fseek(file, 0, SEEK_END);       
    fileSize = ftell(file);
    fseek(file, 0, SEEK_SET);       

    // Round up to 32k, might not be necessary for multiboot
    if(fileSize & (32768 - 1))
        fileSize += 32768;
    fileSize &= ~(32768 - 1);

    // Initialize command buffer
    memset(&sm, 0, sizeof(sm));
    sm.command = CMD_WRITEDATA;
    sm.size = fileSize;
    sm.magic = MAGIC_NUMBER;
    sm.unknown = 0x6; // no idea what this is...
    sm.address = GBA_EWRAM;
    sm.sizekb = fileSize / 1024;

    // Send the data to EWRAM
    kernResult = f2a_sendMessage(&sm);
    if (kernResult != kIOReturnSuccess)
    {
        printf("F2A Error: Unable to initiate Multiboot download (%08x)\n", kernResult);
        return kernResult;
    }

    // Another spot that seems to require some time
    sleep(3);

    UInt32 j;
    for(j = 0; j < fileSize; j += 1024)
    {
        if (gVerbosity > 0)
            printf("Writing Multiboot data %d\n", (int)j);
        
        // Read in 1K of data from the file
        bytesRead = fread(fileBuffer, 1, 1024, file);
        if (ferror(file))
        {
            printf("File Error: Could not read Multiboot file (%d)\n", ferror(file));
            return kIOReturnError;
        }
        
        // Write 1K of data to the EWRAM
        kernResult = usb_write(fileBuffer, 1024);
        if (kernResult != kIOReturnSuccess)
        {
            printf("F2A Error: Unable to write Multiboot data (%08x)\n", kernResult);
            return kernResult;
        }
    }
    
    // Launch the program
    memset(&sm, 0, sizeof(sm));
    sm.command = CMD_WRITEDATA;
    sm.magic = MAGIC_NUMBER;
    sm.unknown = 0x8; // no idea what this is...
    sm.address = GBA_EWRAM;
        
    kernResult = f2a_sendMessage(&sm);
    if (kernResult != kIOReturnSuccess)
    {
        printf("F2A Error: Unable to initiate Multiboot download (%08x)\n", kernResult);
        return kernResult;
    }
    
    return kernResult;
}

IOReturn f2a_writeRoms(UInt8 numFiles, FILE** files, UInt32 address)
{
    IOReturn 	kernResult;
    sendmsg 	sm;
    
    FILE* 	file;
    UInt32 	fileSize;
    char 	fileBuffer[1024];
    size_t 	bytesRead;
		
    int i;
    for (i = 0; i < numFiles; i ++)
    {
    
        file = files[i];

        fseek(file, 0, SEEK_END);       
        fileSize = ftell(file);
        fseek(file, 0, SEEK_SET);       

        // Round up to 32k, FIXME: This has to be 128k for Turbo carts
        if(fileSize & (32768 - 1))
            fileSize += 32768;
        fileSize &= ~(32768 - 1);

        // Initialize command buffer
        memset(&sm, 0, sizeof(sm));
        sm.command = CMD_WRITEDATA;
        sm.size = fileSize;
        sm.magic = MAGIC_NUMBER;
        sm.unknown = 0xa; // no idea what this is...
        sm.address = address;
        sm.sizekb = fileSize / 1024;

        // Flash the ROM
        kernResult = f2a_sendMessage(&sm);
        if (kernResult != kIOReturnSuccess)
        {
            printf("F2A Error: Unable to initiate ROM download (%08x)\n", kernResult);
            return kernResult;
        }

        // Another spot that seems to require some time
        sleep(3);

        UInt32 j;
        for(j = 0; j < fileSize; j += 1024)
        {
            if (gVerbosity > 0)
                printf("Writing ROM data %d\n", (int)j);
            
            // Read in 1K of data from the file
            bytesRead = fread(fileBuffer, 1, 1024, file);
            if (ferror(file))
            {
                printf("File Error: Could not read ROM file (%d)\n", ferror(file));
                return kIOReturnError;
            }
            
            // Write 1K of data to the ROM
            kernResult = usb_write(fileBuffer, 1024);
            if (kernResult != kIOReturnSuccess)
            {
                printf("F2A Error: Unable to write ROM data (%08x)\n", kernResult);
                return kernResult;
            }
        }
        
        // Increment destination address in case we're sending more files
        address += fileSize;
    }
    
    return kernResult;
}

IOReturn f2a_readMemory(FILE* file, UInt32 address, UInt32 size)
{
    IOReturn 	kernResult;
    sendmsg 	sm;
    
    char 	dataBuffer[1024];
    size_t 	bytesWritten;

    // Initialize command buffer
    memset(&sm, 0, sizeof(sm));
    sm.command = CMD_READDATA;
    sm.size = size;
    sm.magic = MAGIC_NUMBER;
    sm.unknown = 0x7; // no idea what this is...
    sm.address = address;
    sm.sizekb = size / 1024;

    // Flash the ROM
    kernResult = f2a_sendMessage(&sm);
    if (kernResult != kIOReturnSuccess)
    {
        printf("F2A Error: Unable to initiate ROM download (%08x)\n", kernResult);
        return kernResult;
    }

    int i;
    for(i = 0; i < sm.sizekb; i++)
    {
            if(usb_read(dataBuffer, 1024)==-1)
            {
                    close(f);
                    return -1;
            }
            
            fwrite(file, data, 1024)
            {
                    perror("unable to write data to file");
                    close(f);
                    return -1;
            }
    }

    return kernResult;
}
