// OS X port (C) 2003 Eli Curtz <eli@nuprometheus.com>
// Original Linux version (C) 2003 Ulrich Hecht <uli@emulinks.de>
// Licensed under the terms of the GNU Public License version 2

#include <unistd.h>

#include <CoreFoundation/CoreFoundation.h>

#include <IOKit/IOKitLib.h>

#include "flash2a.h"


// constants
#define F2A_VERSION	"0.1a"

#define MODE_WRITE	0
#define MODE_READ	1
#define MODE_MULTIBOOT	2

#define MAX_FILES	64

// globals
UInt8		gMode = MODE_WRITE;
UInt8		gVerbosity = 1;
UInt32		gAddress = GBA_ROM;
UInt32		gOffset = 0;
UInt32		gBlocks = 0;

FILE*		gFiles[MAX_FILES];

int main(int argc, const char* argv[])
{
    int option = 0;
    int numFiles = 0;
    int dataSize = 0;
    const char** fileNames = NULL;
    FILE* loaderFile;
    
    do
    {
        option = getopt(argc, (char**)argv, "rwmsb:o:vh");
        
        switch(option)
        {
	case 'r':
            gMode = MODE_READ;
            break;
        case 'w':
            gMode = MODE_WRITE;
            break;
        case 'm':
            gMode = MODE_MULTIBOOT;
            gAddress = GBA_EWRAM;
            break;
        case 's':
            gAddress = GBA_SRAM;
            break;
        case 'b':
            gBlocks = atoi(optarg);
            break;
        case 'o':
            gOffset = atoi(optarg);
            break;
        case 'v':
            gVerbosity++;
            break;
        case 'h':
        case '?':
            printf("\nf2a (OS X) %s\n\n", F2A_VERSION);
            printf("Usage: %s [OPTION]... [FILE]...\n\n", argv[0]);
            printf("  -r\tread data from cart\n");
            printf("  -w\twrite data to cart (default)\n");
            printf("  -m\twrite data to multiboot RAM\n");
            printf("  -s\taccess SRAM (default is to access ROM)\n");
            printf("  -b\tblocks of data to transfer (MBit for ROM, kBit for SRAM)\n");
            printf("  -o\toffset from base address (MBit for ROM, kBit for SRAM)\n");
            printf("  -v\tincrease verbosity\n");
            printf("  -h\tdisplay this help message\n");
            exit(1);
        }
    } while(option != -1);

    // Check file count
    numFiles = argc - optind;
    if (numFiles <= 0)
    {
            printf("%s: file name missing\n", argv[0]);
            exit(1);
    }
    if (gMode == MODE_READ && numFiles != 1)
    {
            printf("%s: cannot read into multiple files\n", argv[0]);
            exit(1);
    }
    if (gMode == MODE_MULTIBOOT && numFiles != 1)
    {
            printf("%s: cannot write multiple multiboot files\n", argv[0]);
            exit(1);
    }
    if (numFiles > MAX_FILES)
    {
        numFiles = MAX_FILES;
        printf("WARNING: cannot write more than %d files\n", MAX_FILES);
    }
    
    // Adjust the address
    if (gAddress == GBA_ROM)
    {
        gAddress += 128 * 1024 * gOffset;
        dataSize = 128 * 1024 * gBlocks;
    }
    else if (gAddress == GBA_SRAM)
    {
        gAddress += 128 * gOffset;
        dataSize = 128 * gBlocks;
    }
    
    // Get the files
    fileNames = argv + optind;
    if (gMode == MODE_READ)
    {
        gFiles[0] = fopen(fileNames[0], "wb"); 
        if (gFiles[0] == NULL)
        {
            printf("File Error: could not create %s\n", fileNames[0]);
            exit(1);
        }
    }
    else
    {
        int i;
        for (i = 0; i < numFiles; i++)
        {
            gFiles[i] = NULL;
            gFiles[i] = fopen(fileNames[i], "rb");
            if (gFiles[i] == NULL)
            {
                printf("File Error: could not open %s\n", fileNames[i]);
                while (--i >= 0)
                    fclose(gFiles[i]);
                exit(1);
            }
        }
    }
    
    IOReturn status = EstablishCommunication();
    if (status == kIOReturnSuccess)
    {
        usb_connect();
    
        loaderFile = NULL;
        loaderFile = fopen("loader.gba", "rb");
        if (loaderFile)
        {
            f2a_boot(loaderFile);
            fclose(loaderFile);
        }
        
        if (gMode == MODE_MULTIBOOT)
        {
            f2a_writeMultiboot(gFiles[0]);
            fclose(gFiles[0]);
        }
        else if (gMode == MODE_READ)
        {
            f2a_readMemory(gFiles[0], gAddress, dataSize);
            fclose(gFiles[0]);
        }
        else
        {
            f2a_writeRoms(numFiles, gFiles, gAddress);
            int i;
            for (i = 0; i < numFiles; i++)
            {
                fclose(gFiles[i]);
            }
        }
    
        usb_disconnect();
    }
    else if (status == kIOReturnNoDevice)
    {
        printf("F2A Error: could not locate linker\n");
    }
    else
    {
        printf("F2A Error: could not connect to linker (%08x)\n", status);
    }

    return 0;
}
